#!/bin/bash
#
# uc4_functions.sh
#
# Description
# Functions for UC4 (Automation Engine) administrators
#
# Usage
# Source this script in .bashrc, and then use the functions like smcl and ae.
#
# This script relies on variables set in serverr_.env.sh and uc4_env.sh, so
# source those first. I.e.g
# . server_env.sh
# . uc4_env.sh
# . uc4_functions.sh
#
# Author
# Michael A. Lowry <michael_lowry@swissre.com> <michael.lowry@gmail.com>
#
# Version	Date		Description
# 1.0		2020.08.13	Initial version
# 2.0		2021.03.01	Added ae function to control Automation Engine via the SMgr.
# 2.1		2022.10.13	Added pps function to list child processes.
# 2.2		2024.02.28	Added functions to identify JWPs by role.
# 2.2		2024.03.18	Added aex function. Updated functions that identify AE processes.
# 2.3		2024.03.20	Added aep function. Fixed dwp function.

# Set shell options
shopt -s extglob # Extended globbing enables lists like '@(A|B|C)'

set_environment () {
if [[ "${SERVER_ENVIRONMENT}" == "" ]] || [[ "${SERVER_NODE}" == "" ]]; then
	echo "Please set the following variables and try again:"
	echo 
	echo "Variable			Accepted value(s)"
	echo "SERVER_ENVIORNMENT	UC4_EXP2 | UC4_EXP | UC4_DEV | UC4_ITE | UC4_PROD"
	echo "SERVER_NODE		A | B"
	echo
fi
if [[ "$AE_SMGR_HOME}" == "" ]] || [[ "${ORACLE_HOME}" == "" ]]; then
	echo "Please set the following variables and try again:"
	echo 
	echo "Variable			Description"
	echo "AE_SMGR_HOME		Path to Service Manager for the Automation Engine"
	echo "ORACLE_HOME		Path to Oracle client."
	echo
fi
case "${SERVER_FUNCTION}" in
UC4_AE)
	AE_ENV="$SERVER_ENVIRONMENT"
	AE_NODE="$SERVER_NODE"
	# Set INI file name and Service Manager destination name based on details of the server.
	SMgr_INI="$AE_SMGR_HOME/ucybsmgr_server_${AE_ENV}-${AE_NODE}.ini"
	SMgr_Dest="uc4s_${AE_ENV}-${AE_NODE}"
;;
*)
	echo "ERROR: Host ${SMgr_Host} is not a known Automation Engine server."
	echo "Please check the definitions in the /opt/uc4/share/scripts/.server_def script."
	# Set default values for the INI file and Service Manager destination name
	SMgr_INI="$AE_SMGR_HOME/ucybsmgr.ini"
	SMgr_Dest="uc4s"
esac
} # End set_environment function

read_and_validate_files () {
# Validate destination name
smgr_dest_re='^[A-Za-z0-9\_\-]+$'
if [[ ! "${SMgr_Dest}" =~ $smgr_dest_re ]]; then
        echo "ERROR: Destination name contains invalid characters. Valid characters: A-Z, a-z, 0-9, -, _"
	echo "SMgr_Dest: ${SMgr_Dest}"
	return 1
elif [[ ${#SMgr_Dest} -gt 31 ]]; then
        echo "ERROR: Destination name must be 1-31 characters long (1-16 characters for use in the AWI)."
	echo "SMgr_Dest: ${SMgr_Dest}"
	return 1
fi

# Validate length of destination name
if [[ ${#SMgr_Dest} -gt 16 ]] && [[ ${#SMgr_Dest} -lt 31 ]]; then
	echo "WARNING: Destination name is valid for the Service Manager, but it will not work with the"
	echo "Automation Engine. Please use a destination name between 1 and 16 characters long."
	return 1
fi

# Validate existence INI file
if [[ ! -f "${SMgr_INI}" ]]; then echo "INI file ${SMgr_INI} not found."; return 1; fi

# Find passwords in INI file.
Password_L1_set=$(cat ${SMgr_INI} | grep "\S" | grep -v '^;' | grep "^password_l1" | wc -l)
Password_L2_set=$(cat ${SMgr_INI} | grep "\S" | grep -v '^;' | grep "^password_l2" | wc -l)
Password_L3_set=$(cat ${SMgr_INI} | grep "\S" | grep -v '^;' | grep "^password_l3" | wc -l)
if [[ "$Password_L1_set" -eq 1 ]]; then
	Lowest_SMgr_PW_Level=1
elif [[ "$Password_L2_set" -eq 1 ]]; then
	Lowest_SMgr_PW_Level=2
elif [[ "$Password_L3_set" -eq 1 ]]; then
	Lowest_SMgr_PW_Level=3
else
	Lowest_SMgr_PW_Level=0
fi

# Validate existence of destination name in INI file
grep "\[Destination ${SMgr_Dest}\]" ${SMgr_INI} >/dev/null 2>&1
if [[ $? -ne 0 ]]; then echo "Destination ${SMgr_Dest} not found in INI file ${SMgr_INI}."; return 1;fi
# Check for existence of INI file.
if [[ ! -s "${SMgr_INI}" ]]; then
        echo "ERROR: INI file ${SMgr_INI} not found."
fi

# Validate INI, SMD, and SMC files
error="false"

# Find destination name in INI file.
DEST_HIT_COUNT=$(cat ${SMgr_INI} | grep "\S" | grep -v '^;' | grep "\[Destination .*$SMgr_Dest\]" | wc -l)
case "${DEST_HIT_COUNT}" in
0)
        echo "ERROR: Destination ${SMgr_Dest} not found in INI file."
	error="true"
	;;
1)
	if [[ "${VERBOSE_MODE}" == "true" ]]; then
        	echo "ERROR: Destination ${SMgr_Dest} found in INI file."
	fi
	;;
*)
        echo "ERROR: Destination ${SMgr_Dest} found more than once in INI file."
	error="true"
	;;
esac

# Get SMD & SMC file paths from INI file
SMD_FILE=$(cat ${SMgr_INI} | grep "\S" | grep -v '^;' | grep -A2 "\[Destination .*$SMgr_Dest\]" | awk -F= '$1=="deffile" {print $2}')
SMC_FILE=$(cat ${SMgr_INI} | grep "\S" | grep -v '^;' | grep -A2 "\[Destination .*$SMgr_Dest\]" | awk -F= '$1=="cmdfile" {print $2}')

# Check for SMD file in destination specification.
if [[ "$SMD_FILE" == "" ]]; then 
        echo "ERROR: SMD file not specified in destination."
	error="true"
fi
# Check for SMC file in destination specification.
if [[ "$SMC_FILE" == "" ]]; then 
        echo "ERROR: SMC file not specified in destination."
	error="true"
fi

# Check for existence of SMD file.
if [[ ! -s "${SMD_FILE}" ]]; then
        echo "ERROR: Specified SMD file does not exist."
	error="true"
fi
# Check for existence of SMC file.
if [[ ! -s "${SMC_FILE}" ]]; then
        echo "ERROR: Specified SMC file does not exist."
	error="true"
fi

# Find SMgr port number in INI file.
SMgr_Port=$(awk -F= '$1 ~ /port/ {print $2}' ${SMgr_INI})
# Validate SMgr port number
if [[ "${SMgr_Port}" == "" ]]; then
        echo "ERROR: No SMgr port number found in INI file."
	error="true"
fi
re='^[0-9]+$'
if [[ ! "${SMgr_Port}" =~ $re ]] || [[ "${SMgr_Port}" -gt 65535 ]]; then
        echo "ERROR: Invalid SMgr port number '${SMgr_Port}' found in INI file."
	error="true"
fi

if [[ "${error}" == "true" ]]; then return 1;fi
} # End read_and_validate_files function

read_pw () {
Command_PW_Level="$1"
if [[ "${SMgr_Pass}" == "" ]]; then
	if [[ "${VERBOSE_MODE}" == "true" ]]; then
		echo "Service Manager password not set yet."
		echo "Lowest Service Manager Password Level: ${Lowest_SMgr_PW_Level}"
		case "${Lowest_SMgr_PW_Level}" in
		1)
			echo "Password required to get process list."
			;;
		2)
			echo "Password required only to make changes."
			;;
		3)
			echo "Password required only for administration."
			;;
		0)
			echo "Password not required."
			;;
		*)
			echo "Unexpected value of Lowest_SMgr_PW_Level variable."
			;;
		esac
	fi
	if [[ "${Command_PW_Level}" -ge "${Lowest_SMgr_PW_Level}" ]] && [[ "${Lowest_SMgr_PW_Level}" -ne 0 ]]; then
		echo "Service Manager password required to perform this operation."
		read -s -p "Password: " SMgr_Pass
		echo
		if [[ "${SMgr_Pass}" == "" ]]; then
			echo "ERROR: Empty password."
		else	
			SMgr_Pass_String=" -p ${SMgr_Pass}"
		fi
	else
		if [[ "${VERBOSE_MODE}" == "true" ]]; then
			echo "Service Manager password not required to peform this operation."
		fi
	fi
fi
}

smcl_usage(){
cat <<EOF
################################################################################
#####                                 smcl                                 #####
################################################################################
Wrapper script for the Service Manager Command Line program ucybsmcl,
adapted for use with the Service Managers that control the Automation Engine
systems at Swiss Re.

Usage: 

Simple syntax:
smcl [list]			: List services (AE server processes).
smcl start <service>		: Start a service.
smcl stop <service>		: Stop a service.
smcl shutdown <service>		: Shut down the AE system (via a WP).

Classic syntax:
smcl [ -c GET_PROCESS_LIST ]	: List services (AE server processes). 
smcl -c START_PROCESS ...	: Start a service.
smcl -c STOP_PROCESS ...	: Stop a service.
smcl -c SET_DATA ...		: Change the properties of a service.

It is not necessary to include the '-h <hostname:port>' & '-n <phrase>' options
normally required by ucybsmcl. The smcl function automatically fills in these
details, based on the server where you run it.

It also unnecessary to include the '-p <password>' option. If a password is
required, you will be prompted for it.

See /opt/uc4/share/scripts/uc4_env.sh for more details about how smcl works.
See the doc page for ucybsmcl for details about the other options (...) above.

Man page: https://wiki.swissre.com/pages/viewpage.action?pageId=401903590
################################################################################

EOF
}

smcl () {
# Wrapper function for the Service Manager Command Line program ucybsmcl.
# This wrapper fills in the host, port and destination (phrase), sends error messages to
# /dev/null, and prints tab-delimited output for GET_PROCESS_LIST commands.
# TODO store pw in file & nohup sleep rm to remove it after an hour?
# Remove pw file also if error 3 is returned. 
smcl_args="$@"
if [[ "${SMgr_INI}" == "" ]]; then set_environment; fi
if [[ "${SMgr_Port}" == "" ]]; then read_and_validate_files; fi
ucybsmcl_out="/tmp/ucybsmcl_out_$$.txt"
ucybsmcl_err="/tmp/ucybsmcl_err_$$.txt"
# Regular expressions for matching SMgr sub-commands
re1='^.*GET_PROCESS_LIST.*$'
re2='^.*START_PROCESS.*$'
re3='^.*STOP_PROCESS.*$'
re4='^.*SET_DATA.*$'
# Regular expression for matching SMgr service names
re5='^[A-Za-z0-9\_\-]+$'
# Default: get list of running processes
if [[ "$smcl_args" == "" ]]; then
	smcl_args="-c GET_PROCESS_LIST"
	smcl_header="Service\t\tStatus\tPID\tStart time\t\tCPU time\n"
fi
arg1=$1
arg2=$2
case $arg1 in
list)
	smcl_args="-c GET_PROCESS_LIST"
	;;
start)
	smcl_args="-c START_PROCESS"
	;;
stop)
	smcl_args="-c STOP_PROCESS -m C"
	;;
shutdown)
	smcl_args="-c STOP_PROCESS -m S"
	;;
help)
	smcl_usage
	return 1
	;;
esac
case $arg1 in @(start|stop|shutdown))
	if [[ ! "$arg2" =~ $re5 ]]; then
		echo "ERROR: '$arg2' does not look like a valid service name."
		smcl_usage
		return 1
	else
		smcl_args2="-s ${arg2}"
	fi
	;;
esac
if [[ "$smcl_args" =~ $re1 ]]; then
	read_pw 1
	smcl_cmd="GET_PROCESS_LIST"
	if [[ "${smcl_header}" != "" ]]; then printf "${smcl_header}"; fi
elif [[ "$smcl_args" =~ $re2 ]] || [[ "$smcl_args" =~ $re3 ]] || [[ "$smcl_args" =~ $re4 ]]; then
	read_pw 2
	if [[ "$smcl_args" =~ $re2 ]]; then smcl_cmd="START_PROCESS"; fi
	if [[ "$smcl_args" =~ $re3 ]]; then smcl_cmd="STOP_PROCESS"; fi
	if [[ "$smcl_args" =~ $re4 ]]; then smcl_cmd="SET_DATA"; fi
else
	echo "ERROR: Invalid smcl command."
	echo
	smcl_usage
	return 1
fi
ucybsmcl_command="${AE_SMGR_HOME}/ucybsmcl $smcl_args -h ${SMgr_Host}:${SMgr_Port} -n ${SMgr_Dest}${SMgr_Pass_String} ${smcl_args2} 1>${ucybsmcl_out} 2>${ucybsmcl_err}"
${AE_SMGR_HOME}/ucybsmcl $smcl_args -h ${SMgr_Host}:${SMgr_Port} -n ${SMgr_Dest}${SMgr_Pass_String} ${smcl_args2} 1>${ucybsmcl_out} 2>${ucybsmcl_err}
ucybsmcl_rc=$?
case $ucybsmcl_rc in
0) ucybsmcl_message="A command has been processed without an error.";;
1) ucybsmcl_message="An incorrect command has been entered.";;
2) ucybsmcl_message="No active ServiceManager has been located on the specified computer.";;
   # Possible reason: Either the name of the computer was not entered, or no ServiceManager is active on this computer.";;
3) ucybsmcl_message="The ServiceManager reacts to a command in an unexpected way."
   # Possible reason: The ServiceManager version is not up to date.";;
   # Also wrong password.
	if [[ "${SMgr_Pass}" != "" ]]; then unset SMgr_Pass; fi;;
4) ucybsmcl_message="Errors occurred with the ServiceManager connection.";;
5) ucybsmcl_message="The specified ServiceManager environment is not present on the corresponding computer.";;
   # Possible reason: The ServiceManager for the specified ServiceManager environment is either not installed or not started.";;
*) ucybsmcl_message="Other error.";;
esac
if [[ ${ucybsmcl_rc} -eq 0 ]]; then
	if [[ "${smcl_cmd}" == "GET_PROCESS_LIST" ]]; then
		# For GET_PROCESS_LIST, check for '*** Access denied ***' string in output
		if [[ $(grep "*** Access denied ***" "${ucybsmcl_out}" | wc -l ) -ne 0 ]]; then
			echo "ERROR: 'Access denied' error returned by Service Manager."
			ucybsmcl_rc=9 # Not an actual ucybsmcl return code.
			ucybsmcl_message="Access denied."
		else 
			cat ${ucybsmcl_out} | sed -e 's/" "/\t/g' | sed -e 's/^.\(.*\).$/\1/'
		fi
	else
		cat ${ucybsmcl_out}
	fi
else # Non-zero exit code.
	echo "Command: ${ucybsmcl_command}"
	smcl_usage
	echo "ucybsmcl RC ${ucybsmcl_rc}: ${ucybsmcl_message}"
	cat "${ucybsmcl_out}"
	if [[ "${VERBOSE_MODE}" == "true" ]]; then
		cat "${ucybsmcl_err}" >/dev/stderr
	fi
fi
if [[ -f "${ucybsmcl_out}" ]]; then rm "${ucybsmcl_out}"; fi
if [[ -f "${ucybsmcl_err}" ]]; then rm "${ucybsmcl_err}"; fi
return "${ucybsmcl_rc}"
} # End ucybsmcl function

get_ae_server_smgr_details () {
# Read details of running AE Server Service Manager process, including PID, INI file, and destination name.
SMgr_PID_count=$(  ps -ef | grep "${AE_SMGR_HOME}" | grep -v grep | wc -l )
if [[ "${SMgr_PID_count}" -gt 1 ]]; then echo "More than one AE Server Service Manager running!"; return 1; fi
SMgr_PID=$( ps -ef | grep "${AE_SMGR_HOME}" | awk -v smgrhome="${AE_SMGR_HOME}" '$8 ~ smgrhome && $8 !~ /grep/ {print $2}' )
SMgr_Proc_INI=$( ps -ef | grep "${AE_SMGR_HOME}" | awk -v smgrhome="${AE_SMGR_HOME}" '$8 ~ smgrhome && $8 !~ /grep/ {print $9}' | cut -c3- )
if [[ "${SMgr_Proc_INI:(-4)}"  == ".ini" ]]; then SMgr_INI="${SMgr_Proc_INI}"; fi
SMgr_Proc_Dest=$( ps -ef | grep "${AE_SMGR_HOME}" | awk -v smgrhome="${AE_SMGR_HOME}" '$8 ~ smgrhome && $8 !~ /grep/ {print $10}' )
smgr_dest_re='^[A-Za-z0-9\_\-]+$'
if [[ "${SMgr_Proc_Dest}" =~ $smgr_dest_re ]] && [[ ${#SMgr_Proc_Dest} -lt 31 ]]; then SMgr_Dest="${SMgr_Proc_Dest}"; fi
} # End get_ae_server_smgr_details function

read_pw () {
Command_PW_Level="$1"
if [[ "${SMgr_Pass}" == "" ]]; then
	if [[ "${VERBOSE_MODE}" == "true" ]]; then
		echo "Service Manager password not set yet."
		echo "Lowest Service Manager Password Level: ${Lowest_SMgr_PW_Level}"
		case "${Lowest_SMgr_PW_Level}" in
		1)
			echo "Password required to get process list."
			;;
		2)
			echo "Password required only to make changes."
			;;
		3)
			echo "Password required only for administration."
			;;
		0)
			echo "Password not required."
			;;
		*)
			echo "Unexpected value of Lowest_SMgr_PW_Level variable."
			return 1
			;;
		esac
	fi
	if [[ "${Command_PW_Level}" -ge "${Lowest_SMgr_PW_Level}" ]] && [[ "${Lowest_SMgr_PW_Level}" -ne 0 ]]; then
		echo "Service Manager password required to peform this operation."
		read -s -p "Password: " SMgr_Pass
		echo
		if [[ "${SMgr_Pass}" == "" ]]; then
			echo "ERROR: Empty password."
			return 1
		else	
			SMgr_Pass_String=" -p ${SMgr_Pass}"
		fi
	else
		if [[ "${VERBOSE_MODE}" == "true" ]]; then
			echo "Service Manager password not required to peform this operation."
		fi
	fi
fi
}

get_ae_server_proc_count () {
Running_Service_Count=$(smcl -c GET_PROCESS_LIST | awk -F '\t' '$2 == "R" {print $1,"\t",$2}' | wc -l)
ucybsmcl_RC=${PIPESTATUS[0]}
if [[ "${ucybsmcl_RC}" -ne 0 ]]; then
	echo "ERROR: ucybsmcl return code $ucybsmcl_RC"
	return
fi
} # End get_ae_server_proc_count

ae_server_status () {
# Check for AE Server Service Manager(s) already running.
local action=$1
get_ae_server_smgr_details
if [[ "$action" == "start" ]]; then ALREADY="already "; fi
case "${SMgr_PID_count}" in
0)
	if [[ "$action" != "start" ]]; then
		echo "No AE Server Service Manager running."
	fi
	;;
*) # TODO Add special handling for case when >1 AE SMgr is running or remove support for this altogether.
	if [[ "${SMgr_PID_count}" -gt 1 ]]; then
		PLURAL_SUFFIX="s"
		NUMBER_PREFIX="${SMgr_PID_count} "
	fi
	echo "${NUMBER_PREFIX}AE Server Service Manager${PLURAL_SUFFIX} ${ALREADY}running."
	ps -ef | egrep "${AE_SMGR_HOME}|STIME" | grep -v grep
	SMgr_child_processes=$( ps -ef | awk -v pid="${SMgr_PID}" '$3 == pid' | grep -v ucybsmgr-listener | wc -l )
	echo "Running child processes: ${SMgr_child_processes}"
	
	;;
esac	
} # End ae_server_status function

ae_server_list () {
smcl -c GET_PROCESS_LIST
ucybsmcl_RC=$?
if [[ "${ucybsmcl_RC}" -ne 0 ]]; then
	echo "ERROR: ucybsmcl return code $ucybsmcl_RC"
	return
fi
} # End ae_server_list function

wait_for_ae_processes_to_end () {
echo "Stop commands sent successfully. Waiting for processes to end."
# Wait for AE Server processes to end.
local max_retries=10
local retry_delay=5
counter=0
sleep ${retry_delay}
while [[ $counter -le $max_retries ]]; do
        (( counter++ ))
	get_ae_server_proc_count
        if [[ "${Running_Service_Count}" -gt 0 ]]; then
                if [[ $counter -ge $max_retries ]]; then
                        echo "ERROR: ${Running_Service_Count} AE Server processes still running. Giving up."
                        return 1
                else
                        sleep ${retry_delay}
			if [[ $counter -eq 1 ]]; then
                        	echo "WARNING: ${Running_Service_Count} AE Server processes still running. Will check again up to ${max_retries} times at ${retry_delay}-second intervals."
			else
                        	echo "         ${Running_Service_Count} AE Server processes still running. Waiting ${retry_delay} seconds."
			fi
                fi
        else
                break
        fi
done
}

ae_smgr_kill () {
if [[ "${SMgr_PID_count}" -eq 0 ]]; then
	echo "Nothing to do."
	return 1
fi
echo "Are you sure you wish to kill the AE Server Service Manager? This will kill any running child proceses."
select yn in "Yes" "No"; do
case $yn in
	Yes )break;;
	No ) echo "Kill cancelled.";return;;
	esac
done
echo "Killing AE Server Service Manager process with PID ${SMgr_PID}."
kill -TERM ${SMgr_PID}
if [[ $? -eq 0 ]]; then
	ps -ef | egrep "${AE_SMGR_HOME}|STIME" | grep -v grep
else
	echo "Error running kill command."
fi
} # End ae_server_kill function

ae_server_stop () {
if [[ "${SMgr_PID_count}" -gt 1 ]]; then
	return 1
fi
# Test to make sure ucybsmcl returns a valid list.
get_ae_server_proc_count
if [[ "${Running_Service_Count}" -eq 0 ]]; then
	echo "No running AE Server processes. Nothing to do."
	return
fi 
echo "Are you sure you wish to stop all AE Server processes on this node?"
echo "If you choose 'Yes', the STOP_PROCESS command (stop mode C) will be sent to all running AE Server processes on this node."
select yn in "Yes" "No"; do
case $yn in
	Yes )break;;
	No ) echo "Shutdown cancelled.";return;;
	esac
done

echo "Stopping AE Server processes."
while read -r AE_Server_Process; do
echo "Running command :${AE_SMGR_HOME}/ucybsmcl -c STOP_PROCESS -s ${AE_Server_Process} -m C -h ${SMgr_Host}:${SMgr_Port} -n ${SMgr_Dest} -p XXXXXXXX"
# Send stop mode C to each running AE server process
smcl -c STOP_PROCESS -s ${AE_Server_Process} -m C
done <<<"$( smcl -c GET_PROCESS_LIST | awk '$2 == "R" {print $1}')"
wait_for_ae_processes_to_end
echo "All AE server processes stopped successfully."
} # End ae_server_stop function

UC4_Log_Dir="/var/uc4/server"

log2proc () {
Log_Basename=$(basename "$1")
Proc_Type="${Log_Basename:0:2}"
Proc_Num="$(echo ${Log_Basename} | cut -d'_' -f 4)"
Proc_Name="${Proc_Type}${Proc_Num}"
echo "${Proc_Name}"
}

normalizeProcName () {
AE_Proc_Name=$1
AE_Proc_Name=$(echo "$AE_Proc_Name" | tr [:lower:] [:upper:])
Proc_Type="${AE_Proc_Name:0:2}"
Proc_Num="${AE_Proc_Name:2}"
case $Proc_Type in @(CP|WP));; *) echo "ERROR: Process type must be CP or WP." >&2;; esac
if   [[ ${#Proc_Num} -eq 1 ]]; then Proc_Num="00$Proc_Num";
elif [[ ${#Proc_Num} -eq 2 ]]; then Proc_Num="0$Proc_Num";
fi
AE_Proc_Name="${Proc_Type}${Proc_Num}"
echo "${AE_Proc_Name}"
}

proc2log () {
AE_Proc_Name=$1
Proc_Type="${AE_Proc_Name:0:2}"
Proc_Num="${AE_Proc_Name:2}"
AE_Proc_Log="${UC4_Log_Dir}/${Proc_Type}srv_${AE_ENV}_log_${Proc_Num}_00.txt"
echo "${AE_Proc_Log}"
}

log2pid () {
pid="$1"
log=$(ls -l /proc/${pid}/fd | grep /var/uc4/server | cut -d'>' -f 2 | cut -d' ' -f 2)
echo "$log"
}

pid2log () {
log="$1"
pid=$(fuser "$log" 2>/dev/null | tr -d ' ' )
echo "$pid"
}

pwpname () {
PWP_Process_Name=$(find ${UC4_Log_Dir} -name WPsrv_\*_log_???_00.txt -mmin -60 -exec grep U00003475 {} \; | sort | tail -1 | cut -d\' -f2)
echo "${PWP_Process_Name}"
}

pwp () {
PWP_Process_Name=$(find ${UC4_Log_Dir} -name WPsrv_\*_log_???_00.txt -mmin -60 -exec grep U00003475 {} \; | sort | tail -1 | cut -d\' -f2)
PWP_Number=$(echo "${PWP_Process_Name}" | cut -d# -f2 | cut -c3-)
PWP_Log=$(find ${UC4_Log_Dir} -name WPsrv_*_log_${PWP_Number}_00.txt -mmin -480 | tail -1)
if [[ ! "${PWP_Log}" == "" ]]; then echo "${PWP_Log}"; fi
}

owp () {
#OWP_Log=$(find ${UC4_Log_Dir} -name WPsrv_\*_log_???_00.txt -mmin -360 -exec grep -l "processes roles 'O'" {} \; | tail -1 | cut -d: -f1)
OWP_Log=$(find ${UC4_Log_Dir} -name WPsrv_\*_log_???_00.txt -mmin -360 -exec grep -l "U00003344 Server .* assumes the role 'O'" {} \; | tail -1 | cut -d: -f1)
echo "${OWP_Log}"
}

rwp () {
#RWP_Log=$(find ${UC4_Log_Dir} -name WPsrv_\*_log_???_00.txt -mmin -360 -exec grep -l "processes roles 'R'" {} \; | tail -1 | cut -d: -f1)
RWP_Log=$(find ${UC4_Log_Dir} -name WPsrv_\*_log_???_00.txt -mmin -360 -exec grep -l "U00003344 Server .* assumes the role 'R'" {} \; | tail -1 | cut -d: -f1)
echo "${RWP_Log}"
}

dwp () {
if [[ "${WP_Logs}" == "" ]]; then WP_Logs="$(aewp)"; fi
for wplog in ${WP_Logs}; do
	WP_Name=$(log2proc "${wplog}")
	Final_WP_Mode=$(grep "U00003389 Server" ${wplog} | awk -v wp="$WP_Name" -F"'" '$2 ~ wp {print $6}' | tail -1 )
	if [[ "${Final_WP_Mode}" == "DWP" ]]; then
		if [[ $(echo "${DWP_Log}" | wc -l) -eq 0 ]]; then
			DWP_Log="${wplog}\n"
		else
			DWP_Log="${DWP_Log}${wplog}\n"
		fi
	fi
done
printf "${DWP_Log}"
}

restp () {
REST_Log=$(find ${UC4_Log_Dir} -name CPsrv_\*_log_???_00.txt -mmin -360 -exec grep -l "U00003400 Server 'REST API'" {} \; | sort )
echo "${REST_Log}"
}

jcp () {
JCP_Log=$(find ${UC4_Log_Dir} -name CPsrv_\*_log_???_00.txt -mmin -360 -exec grep -l "U00003400 Server 'JCP'" {} \; | sort )
echo "${JCP_Log}"
}

jwp () {
JWP_Log=$(find ${UC4_Log_Dir} -name WPsrv_\*_log_???_00.txt -mtime -5 -exec grep -l "U00003400 Server 'JWP'" {} \; | sort )
echo "${JWP_Log}"
}

aewp () {
WP_Log=$(find ${UC4_Log_Dir} -name WPsrv_\*_log_???_00.txt -mmin -360 -exec grep -L "U00003400 Server 'JWP'" {} \; | sort )
echo "${WP_Log}"
}

aecp () {
CP_Log=$(find ${UC4_Log_Dir} -name CPsrv_\*_log_???_00.txt -mmin -360 -exec grep -L "U00003400 Server 'JCP'" {} \; | sort )
echo "${CP_Log}"
}

jwp-aut () {
JWP_AUT_Log=$(for log in $(jwp); do grep 'U00045395 Assigned roles: \[.*AUT.*\]' $log /dev/null ; done | sort -t':' -k 2 | tail -1 | cut -d':' -f 1)
echo "${JWP_AUT_Log}"
}

jwp-idx () {
JWP_IDX_Log=$(for log in $(jwp); do grep 'U00045395 Assigned roles: \[.*IDX.*\]' $log /dev/null ; done | sort -t':' -k 2 | tail -1 | cut -d':' -f 1)
echo "${JWP_IDX_Log}"
}

jwp-per () {
JWP_PER_Log=$(for log in $(jwp); do grep 'U00045395 Assigned roles: \[.*PER.*\]' $log /dev/null ; done | sort -t':' -k 2 | tail -1 | cut -d':' -f 1)
echo "${JWP_PER_Log}"
}

jwp-utl () {
JWP_UTL_Log=$(for log in $(jwp); do grep 'U00045395 Assigned roles: \[.*UTL.*\]' $log /dev/null ; done | sort -t':' -k 2 | tail -1 | cut -d':' -f 1)
echo "${JWP_UTL_Log}"
}

readcplogs () {
RESTP_Logs="$(restp)"
JCP_Logs="$(jcp)"
CP_Logs="$(aecp)"
}
readwplogs () {
WP_Logs="$(aewp)"
PWP_Logs="$(pwp)"
OWP_Logs="$(owp)"
RWP_Logs="$(rwp)"
DWP_Logs="$(dwp)"
JWP_Logs="$(jwp)"
JWP_AUT_Logs="$(jwp-aut)"
JWP_IDX_Logs="$(jwp-idx)"
JWP_PER_Logs="$(jwp-per)"
JWP_UTL_Logs="$(jwp-utl)"
}

log2proctype () {
if [[ ! "${AE_Proc_Log}" == "" ]]; then
	AE_Proc_Name=$(log2proc "${AE_Proc_Log}")
	AE_Proc_Type=$(grep "U00003400 Server '" "${AE_Proc_Log}" | awk -F"'" '{print $2}')
else
	echo "Could not identify logfile for PID ${SMgr_Svc_PID}." >&2
	# TODO Determine process type using 'ps' command.
fi
# Remove hash symbol (#) from AE process name, if present.
if echo $AE_Proc_Type | grep -q "#"; then AE_Proc_Type=$(echo $AE_Proc_Type | cut -d"#" -f 2); fi
case $AE_Proc_Type in
JWP)
	JWP_Role=""
	if   echo "${JWP_AUT_Logs}" | grep -q "${AE_Proc_Log}"; then JWP_Role="-AUT"
	elif echo "${JWP_IDX_Logs}" | grep -q "${AE_Proc_Log}"; then JWP_Role="-IDX" 
	elif echo "${JWP_PER_Logs}" | grep -q "${AE_Proc_Log}"; then JWP_Role="-PER" 
	elif echo "${JWP_UTL_Logs}" | grep -q "${AE_Proc_Log}"; then JWP_Role="-UTL" 
	fi
      		AE_Proc_Type="$AE_Proc_Type$JWP_Role" 
	;;
WP)
	WP_Role=""
	if   echo "${PWP_Logs}" | grep -q "${AE_Proc_Log}"; then WP_Role="P"
	elif echo "${OWP_Logs}" | grep -q "${AE_Proc_Log}"; then WP_Role="O"
	elif echo "${RWP_Logs}" | grep -q "${AE_Proc_Log}"; then WP_Role="R"
	elif echo "${DWP_Logs}" | grep -q "${AE_Proc_Log}"; then WP_Role="D"
	fi
      		AE_Proc_Type="$WP_Role$AE_Proc_Type" 
	case $AE_Proc_Type in @(PWP|OWP|RWP)) AE_Proc_Type="$AE_Proc_Type*";; esac
	;;
@(REST API|JCP|CP));;
*)
	echo "Could not determine process type for PID ${SMgr_Svc_PID}." >&2
      		AE_Proc_Type="XX" 
esac
}

# Read AE process list from Service Manager, and collect additional information about each process.
aex () {
readcplogs
readwplogs
# Print header
printf "%-20s  %-7s  %-9s  %-8s  %-7s  %-5s  %s\n" "SMgr service name" "AE Proc" "Type/Role" "Status" "PID" "#Conn" "Log file"
ae list | \
while read -r ae_proc; do
	SMgr_Svc_Name=""
	SMgr_Svc_State=""
	SMgr_Svc_PID=""
	AE_Proc_Connections=""
	AE_Proc_Log=""
	AE_Proc_Type=""
	SMgr_Svc_Name=$(echo ${ae_proc} | awk '{print $1}')
	SMgr_Svc_State=$(echo ${ae_proc} | awk '{print $2}')
	case $SMgr_Svc_State in S) SMgr_Svc_State_Name="Stopped" ;; R) SMgr_Svc_State_Name="Running" ;; *) SMgr_Svc_State_Name="Unknown" ;; esac
	SMgr_Svc_PID=$(echo ${ae_proc} | awk '{print $3}')
	if [[ ! "$SMgr_Svc_PID" == "" ]]; then
		AE_Proc_Connections=$(ls -l "/proc/${SMgr_Svc_PID}/fd" | grep 'socket\:\[' | sort -nk 9 | wc -l)
		AE_Proc_Log=$(ls -l /proc/${SMgr_Svc_PID}/fd | grep /var/uc4/server | cut -d'>' -f 2 | cut -d' ' -f 2)
		log2proctype
	fi
	printf "%-20s  %-7s  %-9s  %-8s  %-7s  %-5s  %s\n" $SMgr_Svc_Name $AE_Proc_Name "$AE_Proc_Type" $SMgr_Svc_State_Name $SMgr_Svc_PID "$AE_Proc_Connections" "$AE_Proc_Log"
done
}

# Identify a specified AE server process.
aep () {
AE_Proc_Name=$(normalizeProcName $1)
AE_Proc_Log="$(proc2log ${AE_Proc_Name})"
AE_Proc_Type="-"
SMgr_Svc_PID="-"
SMgr_Svc_Name="-"
SMgr_Svc_State="-"
SMgr_Svc_State_Name="-"
AE_Proc_Connections="-"
case $Proc_Name in CP*) AE_Proc_Type="CP"; readcplogs;; WP*) AE_Proc_Type="WP"; readwplogs;; *) echo "Process type must be CP or WP.";; esac
if [[ ! -f "${AE_Proc_Log}" ]]; then 
	AE_Proc_Log="-"
else
	SMgr_Svc_PID=$(fuser "${AE_Proc_Log}" 2>/dev/null | tr -d ' ' )
	if [[ "$SMgr_Svc_PID" -gt 0 ]]; then
		ae_proc=$(ae list | grep $SMgr_Svc_PID )
		SMgr_Svc_Name=$(echo ${ae_proc} | awk '{print $1}')
		SMgr_Svc_State=$(echo ${ae_proc} | awk '{print $2}')
		case $SMgr_Svc_State in S) SMgr_Svc_State_Name="Stopped" ;; R) SMgr_Svc_State_Name="Running" ;; *) SMgr_Svc_State_Name="Unknown" ;; esac
		AE_Proc_Connections=$(ls -l "/proc/${SMgr_Svc_PID}/fd" | grep 'socket\:\[' | sort -nk 9 | wc -l)
	else
		SMgr_Svc_PID="-"
	fi
	log2proctype
fi
# Print header
printf "%-20s  %-7s  %-9s  %-8s  %-7s  %-5s  %s\n" "SMgr service name" "AE Proc" "Type/Role" "Status" "PID" "#Conn" "Log file"
printf "%-20s  %-7s  %-9s  %-8s  %-7s  %-5s  %s\n" $SMgr_Svc_Name $AE_Proc_Name "$AE_Proc_Type" $SMgr_Svc_State_Name $SMgr_Svc_PID "$AE_Proc_Connections" "$AE_Proc_Log"
}

ae_server_shutdown () {
if [[ "${SMgr_PID_count}" -gt 1 ]]; then
	return 1
fi
echo "${SMgr_Host}:${SMgr_Port} ${SMgr_Dest}"
echo "Attempting to identify Primary Work Process."
smcl -c GET_PROCESS_LIST | awk -F'\t' '$1 ~ "WP" && $1 !~ "JWP" && $2 == "R"' | sort -k 7 
PWP_Service_Name=$( smcl -c GET_PROCESS_LIST | awk -F'\t' '$1 ~ "WP" && $1 !~ "JWP" && $2 == "R"' | sort -k 7 | tail -1 | cut -f 1)
ucybsmcl_RC=${PIPESTATUS[0]}
if [[ "${ucybsmcl_RC}" -ne 0 ]]; then
	echo "ERROR: ucybsmcl return code $ucybsmcl_RC"
	return
fi
# Double-check the service name.
re1='^[A-Za-z0-9\_\-]+$'
re2='^.*WP.*$'
if [[ ! "${PWP_Service_Name}" =~ $re1 ]] || [[ ! "${PWP_Service_Name}" =~ $re2 ]] ; then
	echo "ERROR: '${PWP_Service_Name}' does not look like a WP service name."
	return 1
fi
echo "Likely PWP, based on name & CPU time: ${PWP_Service_Name}."
get_ae_server_proc_count
echo "Are you sure you wish to shut down the AE Server?"
echo "If you choose 'Yes', the STOP_PROCESS command (stop mode S) will be sent to ${PWP_Service_Name}."
echo "This will completely shut down the AE server, stopping all AE server processes on all nodes in the cluster."
select yn in "Yes" "No"; do
case $yn in
	Yes )break;;
	No ) echo "Shutdown cancelled.";return;;
	esac
done
echo "Shutting down AE Server."
echo "Running AE Server processes (on this node) before shutdown: ${Running_Service_Count}"
smcl -c STOP_PROCESS -s ${PWP_Service_Name} -m S
echo "Shutdown command sent successfully. Waiting for proceses to end."
wait_for_ae_processes_to_end
echo "Shutdown of AE Server complete."
} # End ae_server_shutdown function

check_for_open_sockets () {
local Port=$1
local SearchString=$2
local retry_delay=$3
local max_retries=$4
# Check for sockets on port used by SMgr
counter=0
while [[ $counter -le $max_retries ]]; do
        (( counter++ ))
        netstat -an | egrep "${SearchString}" | grep -l ${Port} >/dev/null 2>&1
        RC=$?
        if [[ $RC -eq 0 ]]; then
                if [[ $counter -ge $max_retries ]]; then
                        echo "ERROR: Socket still found on port ${Port} after ${max_retries} retries. Giving up."
                        return 1
                else
			if [[ $counter -eq 1 ]]; then
                        	echo "WARNING: Socket matching '${SearchString}' found on port ${Port}. Will retry ${max_retries} times at ${retry_delay}-second intervals."
			else
                        	echo "         Socket still found on port ${Port}. Waiting ${retry_delay} seconds."
			fi
                        sleep ${retry_delay}
                fi
        else
                break
        fi
done
echo
} # End check_for_open_sockets function

parse_start_mode () {
local START_MODE=$1
if [[ "${START_MODE}" == "" ]]; then
	echo "No start mode specified. Defaulting to 'normal'."
	START_MODE="normal"
fi
case $START_MODE in
mini*)
	START_MESSAGE="Starting MINIMAL AE server with just one process of each type."
	START_MODE_DESC="Minimal AE server"
	SMgr_Dest_SUFFIX="_${START_MODE}"
	;;
stop*)
	START_MESSAGE="Starting Service Manager without starting AE server processes."
	START_MODE_DESC="No automatic start"
	SMgr_Dest_SUFFIX="_${START_MODE}"
	;;
test*) # For testing destination names.
	START_MESSAGE="Starting Service Manager without starting AE server processes."
	START_MODE_DESC="Testing"
	SMgr_Dest="${START_MODE}"
	;;
normal)
	START_MESSAGE="Starting AE server normally."
	START_MODE_DESC="Normal"
	SMgr_Dest_SUFFIX=""
	;;	
*)
	echo "Unknown AE server start mode. Valid modes: normal, nostart, minimal"
	return 1
	;;	
esac
if [[ "${SMgr_Dest_SUFFIX}" != "" ]]; then SMgr_Dest="${SMgr_Dest}${SMgr_Dest_SUFFIX}";fi
echo "Start mode       : $START_MODE_DESC (${START_MODE})"
echo "Destination      : $SMgr_Dest"
} # End parse_start_mode function

ae_server_start () {
if [[ "${SMgr_PID_count}" -ge 1 ]]; then echo "Aborting."; return 1; fi
PWD_ORIG="$PWD"
cd $AE_SMGR_HOME

check_for_open_sockets "${SMgr_Port}" "CLOSE_WAIT|TIME_WAIT" "${retry_delay}" "${max_retries}"

# Prepare to write new log file.
INI_Stub="${SMgr_INI%.ini}"
export LOG_FILE="${INI_Stub}_${SMgr_Dest}_out.txt"
if [[ -f "$LOG_FILE" ]]; then rm "$LOG_FILE"; fi

# Start Service Manager for Automation Engine
echo "Starting Service Manager for Automation Engine (UC4) server."
echo "$START_MESSAGE"
echo "Running command: nohup $AE_SMGR_HOME/ucybsmgr -i${SMgr_INI} ${SMgr_Dest} >$LOG_FILE 2>&1 &"
nohup $AE_SMGR_HOME/ucybsmgr -i${SMgr_INI} ${SMgr_Dest} >$LOG_FILE 2>&1 &
ps -ef | egrep "$AE_SMGR_HOME/ucybsmgr|STIME" | grep -v grep
cd "${PWD_ORIG}"
return 0
} # End ae_server_start function

ae_usage () {
cat <<EOF
################################################################################
#####                                  ae                                  #####
################################################################################
Bash function for controlling the Automation Engine server & its Service Manager
Author: Michael A. Lowry <michael_lowry@swissre.com> <michael.lowry@gmail.com>

Usage:
ae <action> [start mode]
 action     : status | list | start | stop | shutdown | kill
 start mode : normal | mini | stop

Actions:
 status     : Display status of the AE server Service Manager & # of child processes. (default)
 list       : List all AE server processes under the control of the Service Manager.
 start      : Start the AE server Service Manager, with an optional start mode.
 stop       : Stop all AE server processes on this node.
 shutdown   : Stop all AE server processes on both nodes.
 kill       : Kill the AE server Service Manager & any running child processes.

Start modes:
 For action 'start', an optional start mode may be specified.
 This specifies the behavior after the AE Server Service Manager is started.
 normal     : Start up the AE Server normally, using the default SMC file. (default)
 mini       : Start up the AE Server in minimal mode, with just one process of each type.
 stop       : Start up the AE Server Service Manager, but do not start any child processes.

See /opt/uc4/share/scripts/uc4_env.sh for more details about how ae works.

Man page: https://wiki.swissre.com/pages/viewpage.action?pageId=371690146
################################################################################

EOF
} # End ae_usage function

print_header () {
echo "Host name        : $SMgr_Host"
echo "Environment      : $AE_ENV"
echo "Cluster node     : $AE_NODE"
if [[ $VERBOSE_MODE = "true" ]]; then
echo "Working directory: $(pwd)"
echo "PATH             : $PATH"
echo "LD_LIBRARY_PATH  : $LD_LIBRARY_PATH"
echo "JAVA_HOME        : $JAVA_HOME"
echo "CLASSPATH        : $CLASSPATH"
echo "Log file         : $LOG_FILE"
fi
echo "INI file         : $SMgr_INI"
echo "Destination      : $SMgr_Dest"
echo "SMD file         : $SMD_FILE"
echo "SMC file         : $SMC_FILE"
echo "SMgr port number : $SMgr_Port"
echo
} # End print_header function

ae () { # From old AE_Startup.sh and ae_server.sh scripts
if [[ $# -gt 2 ]]; then ae_usage;return 1; fi
# Start main 
ACTION=$1
MODE=$2
#echo "Action: $ACTION; Mode: $MODE"
set_environment
if [ -z ${1+x} ]; then ACTION="status"; fi
case "$ACTION" in
status)
	if [[ $VERBOSE_MODE = "true" ]]; then echo "Status"; fi
	#read_and_validate_files # Required only if we get child process list via ucybsmcl
	ae_server_status "${ACTION}"
	;;
list)
	if [[ $VERBOSE_MODE = "true" ]]; then echo "List"; fi
	read_and_validate_files
	get_ae_server_smgr_details
	read_pw 1
	ae_server_list
	;;
start)
	if [[ $VERBOSE_MODE = "true" ]]; then echo "Start"; fi
	parse_start_mode "${MODE}"
	if [[ $? -ne 0 ]]; then return $?; fi
	read_and_validate_files
	print_header
	ae_server_status "${ACTION}"
	ae_server_start "${MODE}"
	;;
stop)
	if [[ $VERBOSE_MODE = "true" ]]; then echo "Stop"; fi
	# parse_stop_mode # This function does not exist (yet), but could be added if needed.
	read_and_validate_files
	ae_server_status "${ACTION}"
	read_pw 2
	ae_server_stop #"${MODE}"
	;;
@(shut|shutdown))
	if [[ $VERBOSE_MODE = "true" ]]; then echo "Shutdown"; fi
	read_and_validate_files
	ae_server_status "${ACTION}"
	read_pw 2
	ae_server_shutdown
	;;
kill)
	if [[ $VERBOSE_MODE = "true" ]]; then echo "Kill"; fi
	ae_server_status "${ACTION}"
	ae_smgr_kill
	;;
*)
	ae_usage
	return 1
	;;
esac
} # End ae function

function pps () {
	ppid=$1
	ps -fp $ppid --ppid $ppid
}

